import os
import re
import errno
from operator import methodcaller
from six import PY2
from six.moves import filter

if PY2:
    from StringIO import StringIO
else:
    from io import StringIO


class Settings(object):


    def __init__(self, config):
        self.__config__ = config

    @classmethod
    def subclass(cls, *settings, **kw):
        subclass_dict = {setting.name: setting for setting in settings}
        subclass_dict.update(kw)
        subclass = type(cls.__name__, (cls,), subclass_dict)
        return subclass


class Setting(object):

    def __init__(self, option, **kw):
        self.section = kw.pop('section', None)
        if self.section is None:
            self.section, _, self.option = option.partition(':')
        else:
            self.option = option
        self.name = kw.pop('name', self.option)
        if kw:
            raise ValueError("unexpected keyword args '%s'" % kw.keys())

    def __get__(self, obj, objtype):

        if objtype is None:
            return obj

        config = obj.__config__.load()
        cache = config.__dict__.setdefault('__settings_cache__', {})
        if self in cache:
            return cache[self]

        value = self.get(obj, config)
        cache[self] = value
        return value

    def get(self, settings, config):
        return config.get(self.section, self.option)


class Text(Setting):
    """ A setting that is text. """


class DelimitedList(Setting):
    """ A list of items from a single option. """

    def __init__(self, name, **kwargs):
        super(DelimitedList, self).__init__(name)
        self.delimiter = kwargs.pop('delimiter', '\n')
        self.value_type = kwargs.pop('value_type', None)

    def get(self, settings, config):
        text = config.get(self.section, self.option)
        return list(map(self.value_type, text.split(self.delimiter)))


class Directory(Setting):
    """ A setting that is a directory. """

    def __init__(self, name, **kwargs):
        self.makedirs = kwargs.pop('makedirs', False)
        self.expanduser = kwargs.pop('expanduser', True)
        super(Directory, self).__init__(name, **kwargs)

    def get(self, settings, config):
        dirpath = config.get(self.section, self.option)
        if self.expanduser:
            dirpath = os.path.expanduser(dirpath)
        try:
            if self.makedirs:
                os.makedirs(dirpath)
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise
        return dirpath


class Float(Setting):
    """ A setting that is a floating point number. """

    def get(self, settings, config):
        return config.getfloat(self.section, self.option)


class Integer(Setting):
    """ A setting that is an integer. """

    def get(self, settings, config):
        return config.getint(self.section, self.option)


class Collection(Setting):
    """ Multiple configuration lines combined into one setting. """

    def __init__(self, *settings, **kwargs):
        self.separator = kwargs.pop('separator', '-_')
        self.value_type = kwargs.pop('value_type', None)
        super(Collection, self).__init__(*settings, **kwargs)

    def iteritems(self, config):
        for option, value in config.items(self.section):
            pat = '{0}[{1}](.*)$'.format(self.option, self.separator)
            match = re.match(pat, option)
            if not match:
                continue
            if self.value_type is not None:
                value = self.value_type(value)
            yield match.group(1), value


class List(Collection):
    """ A `Collection` setting that is an list of values.  """

    def get(self, settings, config):
        digit_items = (item for item in self.iteritems(config) if item[0].isdigit())
        sorted_items = sorted(digit_items, key=lambda item: int(item[0]))
        return list(item[1] for item in sorted_items)


class Dict(Collection):
    """ A `Collection` setting that is an dictionary of values.  """

    def get(self, settings, config):
        return dict(self.iteritems(config))


__all__ = [
    "Text",
    "DelimitedList",
    "Directory",
    "Float",
    "Integer",
    "List",
    "Dict",
]



import tempfile
import logging.config
from logging import getLogger
from operator import attrgetter
from ConfigParser import SafeConfigParser
from reconf.settings import Settings
from reconf.loggingconfig import create_logging_config_dict
from reconf.location import (ResourceLocation,
                             FileLocation,
                             EnvironLocation,
                             TestResourceLocation,
                             TestStringLocation)


sort_key = attrgetter('order')


class Config(object):

    def __init__(self):
        self.locations = []
        self.cache = {}
        self.defaults = {'tempdir': tempfile.gettempdir()}

    def add_location(self, location):
        try:
            pos = self.locations.index(location)
            return self.locations[pos]
        except ValueError:
            pass
        self.locations.append(location)
        self.locations.sort(key=sort_key)
        return location

    def add_resource(self, package, resource):
        self.add_location(ResourceLocation(package, resource))

    def add_file(self, path):
        self.add_location(FileLocation(path))

    def add_environ(self, name):
        self.add_location(EnvironLocation(name))

    def add_test_resource(self, package, resource):
        return self.add_location(TestResourceLocation(package, resource))

    def add_test_string(self, string):
        self.add_location(TestStringLocation(string))

    def remove_location(self, location):
        self.locations.remove(location)

    remove = remove_location

    def load(self):
        cache = self.cache
        cache_key = tuple(self.locations)
        if cache_key in cache:
            return cache[cache_key]
        cache.clear()
        config = SafeConfigParser(defaults=self.defaults)

        names = []
        logger = getLogger(__name__)
        for location in self.locations:
            for fp, name in location.files():
                logger.debug("reading %r (%s)", fp, name)
                config.readfp(fp, name)
                names.append(name)

        cache[cache_key] = config
        return config

    def settings(self, *settings, **kwargs):
        return Settings.subclass(*settings, **kwargs)(self)

    def logging_config_dict(self):
        return create_logging_config_dict(self.load())

    def configure_logging(self):
        logging.config.dictConfig(self.logging_config_dict())



from operator import methodcaller
from ast import literal_eval
from ConfigParser import NoSectionError

logging_section = 'logging'
root_section = 'logging.root'


def create_logging_config_dict(cp):

    config_dict = {
        'formatters': create_formatters(cp),
        'filters': create_filters(cp),
        'handlers': create_handlers(cp),
        'loggers': create_loggers(cp),
    }

    for option in options(cp, logging_section):
        if option == 'version':
            config_dict[option] = cp.getint(logging_section, option)
        elif option in ('incremental', 'disable_existing_loggers'):
            config_dict[option] = cp.getboolean(logging_section, option)
        else:
            config_dict[option] = cp.get(logging_section, option)

    if cp.has_section(root_section):
        config_dict['root'] = logger_dict_from_section(cp, root_section)

    if not 'version' in config_dict:
        config_dict['version'] = 1

    return config_dict


def sections_with_prefix(prefix, cp):
    for section in cp.sections():
        if section.startswith(prefix):
            yield section, section[len(prefix):]


def create_formatters(cp):
    formatters = {}
    defaults = {'datefmt': None}
    prefix = 'logging.formatter:'
    for section, formatter_id in sections_with_prefix(prefix, cp):
        formatters[formatter_id] = {
            'format': cp.get(section, 'format', 1, defaults),
            'datefmt': cp.get(section, 'datefmt', 1, defaults),
        }
    return formatters


def create_filters(cp):
    return {}


def create_handlers(cp):
    handlers = {}
    prefix = 'logging.handler:'
    for section, handler_id in sections_with_prefix(prefix, cp):
        handler = handlers[handler_id] = {}
        for option in options(cp, section):
            value = cp.get(section, option)
            if option == 'filters':
                handler[option] = id_list(value)
            elif option in ('class', 'level', 'formatter'):
                handler[option] = value
            else:
                try:
                    handler[option] = literal_eval(value)
                except (SyntaxError, ValueError):
                    handler[option] = value

    return handlers


def create_loggers(cp):
    loggers = {}
    prefix = 'logging.logger:'
    for section, logger_id in sections_with_prefix(prefix, cp):
        loggers[logger_id] = logger_dict_from_section(cp, section)
    return loggers


def logger_dict_from_section(cp, section):
    logger_dict = {}
    for option in options(cp, section):
        if option in ('filters', 'handlers'):
            logger_dict[option] = id_list(cp.get(section, option))
        elif option == 'propagate':
            logger_dict[option] = cp.getboolean(section, option)
        else:
            logger_dict[option] = cp.get(section, option)
    return logger_dict


def options(cp, section):
    """ Get options from section without including defaults. """
    try:
        return [k for k in cp._sections[section].keys() if k != '__name__']
    except KeyError:
        raise NoSectionError(section)


def id_list(value):
    return [s.strip() for s in value.split(',')]



import os
import errno
from logging import getLogger
from contextlib import contextmanager
from StringIO import StringIO
from pkg_resources import resource_stream


def enum(*names, **kw):
    typename = kw.pop('typename', 'Enum')
    values = [(name, i) for i, name in enumerate(names)]
    typedict = dict(values)
    typedict['FIRST'] = values[0][1]
    typedict['LAST'] = values[-1][1]
    return type(typename, (), typedict)


#
# This is the order in which configuration files are read.

LocationOrder = enum('RESOURCE', 'STANDARD', 'ENVIRON', 'TEST')


class Location(object):

    order = LocationOrder.STANDARD

    def __init__(self, name):
        self.name = name

    @contextmanager
    def context(self, config):
        """ Context manager for temporarily adding a config `Location`. """
        config.add_location(self)
        try:
            yield self
        finally:
            config.remove_location(self)

    def __hash__(self):
        return hash(self.__class__) ^ hash(self.name)

    def __eq__(self, other):
        return (getattr(other, '__class__') is self.__class__ and
                self.name == other.name)

    def files(self):
        for path in self.name.split(os.pathsep):
            try:
                with open(os.path.expanduser(path), 'rb') as fp:
                    yield fp, path
            except IOError as exc:
                if exc.errno != errno.ENOENT:
                    raise


class FileLocation(Location):
    pass


class ResourceLocation(Location):

    order = LocationOrder.RESOURCE

    def __init__(self, package, resource):
        super(ResourceLocation, self).__init__(os.pathsep.join([package,
                                                                resource]))

    def files(self):
        package, resource = self.name.split(os.pathsep, 1)
        fp = resource_stream(package, resource)
        return [(fp, self.name)]


class EnvironLocation(Location):

    order = LocationOrder.ENVIRON

    def files(self):
        value = os.environ.get(self.name, '')
        for path in value.split(os.pathsep):
            if path.strip() and os.path.exists(path.strip()):
                with open(path, 'rb') as fp:
                    yield fp, path


class TestResourceLocation(ResourceLocation):

    order = LocationOrder.TEST


class TestStringLocation(Location):

    order = LocationOrder.TEST

    def files(self):
        return [(StringIO(self.name), self.name)]


__all__ = [
    'FileLocation',
    'ResourceLocation',
    'EnvironLocation',
    'TestResourceLocation',
]



from setuptools import setup

version = '0.1.3'

setup(
    name='reconf',
    description="Reloadable configuration settings.",
    version=version,
    author='Giles Brown',
    author_email='giles_brown@hotmail.com',
    url='http://github.com/gilesbrown/reconf',
    download_url = 'http://github.com/gilesbrown/reconf/tarball/' + version,
    license='BSD',
    packages=['reconf'],
    install_requires=[
        'six>=0.1.8',
    ],
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        #'Programming Language :: Python :: 3',
        #'Programming Language :: Python :: 3.3',
    ),
)
